import time

from ._internal import defaultarguments, DoreahConfig

_lastcalls = {}


# si		0 means seconds, 1 ms, 2 μs, 3 ns etc
config = DoreahConfig("timing",
	si=0
)


# Take clock. Returns time passed since last call of this function. if called with an identifier, will only
# consider calls with that identifier. No identifier means any call is valid.
# identifiers	arbitrary strings to remember different timers. guaranteed to set all timers to exactly the same time for
#				all identifiers in one call. will return tuple of all identifiers, singular value if only one identifier
def clock(*identifiers):
	"""Takes the time. Returns elapsed time since last call with that identifier for each identifier,
	or time since last call with any identifier if no identifier is given.

	:param string identifiers: Unique identifiers for timing to be taken
	:return: Tuple of elapsed times for each identifier, or single value if only one identifier or no identifier."""

	global _lastcalls

	if len(identifiers) == 0: identifiers = (None,)

	now = time.time()
	# get last calls
	stamps = (_lastcalls.get(i) for i in identifiers)
	results = tuple(None if lc is None else (now - lc) * (1000**config["si"]) for lc in stamps)
	if len(results) == 1: results = results[0]

	# set new stamps
	for i in identifiers:
		_lastcalls[i] = now
	_lastcalls[None] = now			# always save last overall call so we can directly access it

	return results



def clockp(name,*identifiers):
	"""Prints out elapsed time since last call according to the rules of :meth:`clock`

	:param string name: Name to be printed out
	:param string identifiers: Unique identifiers for timing to be taken"""
	time = clock(*identifiers)
	print(name + ": " + str(time))



import pickle
import os
import math
import zlib
import random

from ._internal import DEFAULT, defaultarguments, gopen, DoreahConfig


# folder	folder to store log files
config = DoreahConfig("persistence",
	folder="storage"
)


@defaultarguments(config,folder="folder")
def save(data,name,folder=DEFAULT):
	"""Saves the supplied data structure to disk.

	:param data: Object to be serialized and saved to disk
	:param string name: File name to be used
	:param string folder: Custom folder to save the file in"""

	filename = os.path.join(folder,name)

	try:
		with gopen(filename,"wb") as fl:
			stream = pickle.dumps(data)
			fl.write(stream)
	except:
		pass

@defaultarguments(config,folder="folder")
def load(name,folder=DEFAULT):
	"""Loads a data structure from disk.

	:param string name: File name to be read
	:param string folder: Custom folder where the file is stored
	:return: Data object"""

	filename = os.path.join(folder,name)

	try:
		with gopen(filename,"rb") as fl:
			ob = pickle.loads(fl.read())
	except: ob = None

	return ob

@defaultarguments(config,folder="folder")
def delete(name,folder=DEFAULT):
	"""Deletes a serialized data structure from disk.

	:param string name: File name
	:param string folder: Custom folder where the file is stored"""

	filename = os.path.join(folder,name)
	try:
		os.remove(filename)
	except:
		pass

@defaultarguments(config,folder="folder")
def size(name,folder=DEFAULT):
	"""Returns the size of a serialized object.

	:param string name: File name
	:param string folder: Custom folder where the file is stored"""

	filename = os.path.join(folder,name)
	return os.path.getsize(filename)


# need a deterministic hash function
def _hash(val):
	if isinstance(val,int):
		return val
	elif isinstance(val,str):
		return zlib.adler32(bytes(val,encoding='UTF-8'))
	elif isinstance(val,dict):
		return _hash(tuple((k,val[k]) for k in sorted(val)))
	elif isinstance(val,tuple) or isinstance(val,list):
		x = 10
		i = 1
		for element in val:
			x = ((x * _hash(element)) ** i)
			i += 1
		return x

	try:
		return zlib.adler32(val)
	except: pass
	try:
		return zlib.adler32(bytes(val))
	except: pass



class DiskDict:
	"""Dictionary-like object that stores its authorative information completely on disk.

	:param integer maxmemory: Soft memory limit (in bytes), not meant for precision
	:param integer maxstorage: Soft disk space limit (in bytes), not meant for precision
	:param string name: Directory name used for storage
	:param string folder: Parent directory"""

	@defaultarguments(config,folder="folder")
	def __init__(self,maxmemory=math.inf,maxstorage=1024*1024*1024*4,name="diskdict",folder=DEFAULT):

		self.maxmemory = maxmemory
		self.maxstorage = maxstorage

		self.name = name
		self.folder = os.path.join(folder,name)


		self.cache = {}
		self.cache_unhashable = {}


	def __contains__(self,key):
		try:
			self.get(key)
			return True
		except:
			return False
	def __getitem__(self,key):
		return self.get(key)
	def __setitem__(self,key,value):
		return self.add(key,value)
	def __delitem__(self,key):
		pass

	def __str__(self):
		return "{" + ",".join([str(key) + ":" + str(self.cache[key]) for key in self.cache] + ["etc."]) + "}"

	def get(self,key):
		"""Get the value of a key.

		:param key: Key to be retrieved
		:return: Value of the requested key
		:raises KeyError: No valid entry for the key found."""

		hashvalue = str(_hash(key))
		try:
			return self.cache[key]
		except: pass
		try:
			return self.cache_unhashable[hashvalue]
		except: pass

		if os.path.exists(os.path.join(self.folder,hashvalue)):
			possibilities = os.listdir(os.path.join(self.folder,hashvalue))
			for p in possibilities:
				result = load(os.path.join(self.name,hashvalue,p),folder=self.folder)
				if result[0] == key:
					val = result[1]
					try:
						self.cache[key] = val
					except TypeError:
						self.cache_unhashable[hashvalue] = val
					return val

		raise KeyError()



	def add(self,key,value):
		"""Add an entry.

		:param key: Key to be added
		:param value: Value of this entry"""

		hashvalue = str(_hash(key))

		# cache
		try:
			self.cache[key] = value
		except TypeError:
			self.cache_unhashable[hashvalue] = value

		name = str(random.uniform(1000000000,9999999999)).replace(".","")
		save((key,value),os.path.join(self.name,hashvalue,name),folder=self.folder)



from lxml import etree
from copy import deepcopy
import re
import os

from ._internal import DEFAULT, defaultarguments, DoreahConfig


config = DoreahConfig("pyhp",
	interpret=str
)





@defaultarguments(config,interpret="interpret")
def file(path,d={},interpret=DEFAULT,noroot=False):
	"""Parses a pyhp source file and returns the generated html code.

	:param string path: Location of the pyhp source file
	:param dict d: Variables dictionary
	:param function interpret: Function that determines string representations of variables
	:return: HTML source
	"""

	with open(path,"r") as f:
		content = f.read()

	directory = os.path.dirname(os.path.abspath(path))

	return parse(content,d,interpret=interpret,directory=directory,noroot=noroot)


def _file(path,d,interpret=DEFAULT,noroot=False):

	with open(path,"r") as f:
		content = f.read()

	directory = os.path.dirname(os.path.abspath(path))

	return _parse(content,d,interpret=interpret,directory=directory,noroot=noroot)


@defaultarguments(config,interpret="interpret")
def parse(src,d={},interpret=DEFAULT,directory=None,noroot=False):
	"""Parses pyhp source and returns the generated html code.

	:param string src: Source string
	:param dict d: Variables dictionary
	:param function interpret: Function that determines string representations of variables
	:return: HTML source
	"""

	doc = _parse(src,d,interpret=interpret,directory=directory,noroot=noroot)


	#raw = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
	raw = etree.tostring(doc,encoding="unicode",method="html").replace("&gt;",">").replace("&lt;","<")


	return raw
	# yes I know
	# it's a temporary thing okay

# to tree
def _parse(src,d,interpret=DEFAULT,directory=None,noroot=False):

	if noroot:
		# pseudo root
		src = "<html><body>" + src + "</body></html>"

	doc = etree.HTML(src)
	doc = _parse_node(doc,d,interpret,directory=directory)[0]


	if noroot:
		return list(list(doc)[0]) #all child elements

	return doc


def _parse_node(node,d,interpret,directory=None):


	## fix
	#if node.tag == "html":
	#	node.attrib["xmlns"] = "http://www.w3.org/1999/xhtml"


	## parse pyhp nodes

	if node.tag == "pyhp":

		### CODE

		if len(node.attrib) == 0:
			#print("Executing code!")
			#print({k:d[k] for k in d if not k.startswith("_")})
			code = node.text
			code = code.strip("\t").strip(" ")
			code = code.split("\n")
			if len(code) == 1:
				# one-line code
				code = code[0]
			else:
				# multiline code
				if code[0] != "" or code[-1] != "":
					print("Malformed code block!")
				code = code[1:-1]
				roottabs = 0
				for char in code[0]:
					if char == "\t" or char == " ":
						roottabs += 1
					else:
						break

				code = [line[roottabs:] for line in code]
				code = "\n".join(code)

			exec(code,d)

			return [node.tail]

		#### SAVE

		if _attr(node,"save") is not None and _attr(node,"as") is not None:
			#print("d[" + _attr(node,"as") + "] = " + _attr(node,"save"))
			d[_attr(node,"as")] = eval(_attr(node,"save"),d)
			#print({k:d[k] for k in d if not k.startswith("_")})

			return [node.tail]

		### IMPORT

		if _attr(node,"import") is not None:
			exec("import " + _attr(node,"import"),d)


		#### INCLUDE

		elif _attr(node,"include") is not None:

			filename = _attr(node,"include")
			if directory is None:
				# relative to execution path
				filename = filename
			else:
				# relative to this file
				filename = os.path.join(directory,filename)


			try:
				subnodes = _file(filename,d,interpret=interpret,noroot=True)
				for attr in node.keys():
					if attr != "include":
						subnodes[0].attrib[attr] = _attr(node,attr)
				return subnodes + [node.tail]
			except:
				return [node.tail]


		#### IF

		elif _attr(node,"if") is not None:
			if eval(_attr(node,"if"),d):
				nodestoreturn = [node.text]
				for sn in node:
					nodestoreturn += _parse_node(sn,d,interpret,directory=directory)
				nodestoreturn += [node.tail]
				return nodestoreturn
			else:
				return [node.tail]


		#### FOR IN

		elif _attr(node,"for") is not None and _attr(node,"in") is not None:
			nodestoreturn = []
			# for loop of the elements
			first = True
			try:
				elements = eval(_attr(node,"in"),d)
			except:
				# allow invalid expressions in for loops, just ignore them
				elements = []
			for element in elements:
				if not first and _attr(node,"separator") is not None:
					nodestoreturn += [_attr(node,"separator")]
				first = False

				# in case we overload a dict entry, keep the old one
				sentinel = object()
				if _attr(node,"for") in d:
					hide = d[_attr(node,"for")]
				else:
					hide = sentinel

				# the dict needs to remain the same object so changes from one node
				# in the for loop are carried over into the next loop
				d.update({_attr(node,"for"):element})

				# now go through the nodes each time
				# first add the node text
				nodestoreturn += [node.text]
				# then all the subnodes
				for sn in node:
					sn = deepcopy(sn)
					nodestoreturn += _parse_node(sn,d,interpret,directory=directory)

				# clear the variable after each loop
				del d[_attr(node,"for")]
				if hide is not sentinel:
					d[_attr(node,"for")] = hide


			nodestoreturn += [node.tail]

			return nodestoreturn


		#### ECHO

		elif _attr(node,"echo") is not None:
			#return [d.get(node.get("echo"))]
			return [interpret(eval(_attr(node,"echo"),d))] + [node.tail]

		return []


	## parse normal nodes

	else:

		## replace attributes (not necessary in pyhp nodes)

		for name,value in node.attrib.items():
			vars = re.findall("{.*?}",value)
			for v in vars:
				vname = v[1:-1]
				try:
					value = value.replace(v,interpret(eval(vname,d)))
				except:
					pass
					print("Error parsing:",v,"in attribute")

			node.attrib[name] = value



		subnodes = [n for n in node]
		newsubnodes = []
		for subnode in subnodes:
			node.remove(subnode)
			newsubnodes += _parse_node(subnode,d,interpret,directory=directory)

		prev = None
		for nsn in newsubnodes:


			if nsn is None: continue

			if isinstance(nsn,str):

				# string nodes need to be either the tail of their previous node, or if they're
				# in first position, the text of the parent node
				if prev is None:
					#if node.tag == "head": print(str(nsn) + " appended as text to " + str(node))
					try:
						node.text += nsn
					except:
						node.text = nsn
				else:
					#if node.tag == "head": print(str(nsn) + " appended as tail to " + str(prev))
					try:
						prev.tail += nsn
					except:
						prev.tail = nsn
			else:
				# once we encounter a real node, we can set it as prev for all following text nodes
				prev = nsn



		for nsn in newsubnodes:
			if isinstance(nsn,etree._Element):
				node.append(nsn)


		return [node]



def _attr(node,name):
	res = node.get(name)
	try:
		return res.replace(" gr "," > ").replace(" ls "," < ")
	except:
		return res





### run test server

if __name__ == "__main__":
	from bottle import get, run, static_file
	from doreah.pyhp import file
	import os

	@get("/<path:path>")
	def serve_file(path):

		if os.path.exists(path + ".pyhp"):
			return file(path + ".pyhp")
		if os.path.exists(path):
			return static_file(path,root="")
		if os.path.exists(path + ".html"):
			return static_file(path + ".html",root="")

		return static_file(path,root="") # will produce proper error


	run(host="::",port=1337)



from ._internal import DEFAULT, defaultarguments, DoreahConfig
import zipfile
import os
import urllib.request
import shutil
import distutils.dir_util
import sys

config = DoreahConfig("control")


class Controller:
	def __init__(self,procname,github_project,github_path,custom_functions={},main_script="server.py"):
		self.source_url = "https://github.com/" + github_project
		self.name = github_project.split("/")[-1]
		self.source_path = github_path
		self.custom_functions = custom_functions
		self.main_script = main_script

	def evoke(self):
		if sys.argv[1] == "start": self._start()
		elif sys.argv[1] == "restart": self._start()
		elif sys.argv[1] == "stop": self._stop()
		elif sys.argv[1] == "update": self._update()
		elif sys.argv[1] == "install": _install()
		elif sys.argv[1] in self.custom_functions:
			try:
				self.custom_functions[sys.argv[1]](sys.argv[2:])
			except:
				print("Invalid arguments.")

		else: print("Valid commands:","start","restart","stop","update","install",*[k for k in self.custom_functions])

	def _gotodir(self):
		if os.path.exists(self.main_script):
			return True
		elif os.path.exists("/opt/" + self.name + "/" + self.main_script):
			os.chdir("/opt/" + self.name)
			return True

		print("Installation could not be found.")
		return False

	def _update(self):

		if not self._gotodir(): return False

		if os.path.exists("./.dev"):
			print("Better not overwrite the development server!")
			return

		print("Updating...")

		os.system("wget " + self.source_url + "/archive/master.zip")
		with zipfile.ZipFile("./master.zip","r") as z:


			for f in z.namelist():
				if f.startswith(self.source_path):
					z.extract(f)
			for dir,_,files in os.walk("./" + self.source_path):
				for f in files:
					origfile = os.path.join(dir,f)
					newfile = os.path.join(dir[(len(self.source_path)+1):],f)
					os.renames(origfile,newfile) #also prunes empty directory

		os.remove("./master.zip")


		distutils.dir_util.copy_tree("./" + self.name + "-master/","./",verbose=2)
		shutil.rmtree("./" + self.name + "-master")
		print("Done!")

		os.chmod(self.name,os.stat(self.name).st_mode | stat.S_IXUSR)

		if stop(): start() #stop returns whether it was running before, in which case we restart it


	def generate_bash_completion(self):
		pass



import requests
import mechanicalsoup
import json
import lxml
from lxml import html
import math
import re
import time

from ._internal import DEFAULT, defaultarguments, DoreahConfig

config = DoreahConfig("scraping",
	attempts=10
)




def parse(initial,steps):
	"""Takes a html node and applies all supplied steps to it.

	:param xmlnode initial: Node to start parsing from
	:param list steps: Parsing steps
	:returns: The resulting node or string value
	"""


	i = 0
	result = initial
	has_split = False
	for step in steps:
		#print("Step",step)
		steptype,instruction = step.get("type"),step.get("instruction")
		if not has_split:
			if steptype == "xpath":
				result = result.xpath(instruction)[0]
			elif steptype == "prefix":
				result = instruction + result
			elif steptype == "suffix":
				result = result + instruction
			elif steptype == "rmprefix":
				if result.startswith(instruction): result = result.replace(instruction,"",1)
			elif steptype == "regex":
				result = re.sub(instruction,r"\1",result)
			elif steptype == "last":
				result = result.split(instruction)[-1]


			# these are only applicable for single entries
			elif steptype == "split":
				result = result.split(instruction)
				has_split = True
			elif steptype == "makelist":
				result = [result]
				has_split = True
			elif steptype == "xpathls":
				result = result.xpath(instruction)
				has_split = True
			elif steptype == "follow":
				br = mechanicalsoup.StatefulBrowser()
				page = br.open(result)
				result = html.fromstring(page.content)

		else:
			if steptype == "pick":
				result = result[instruction]
				has_split = False
			elif steptype == "combine":
				result = instruction.join(result)
				has_split = False

			else:

				for k in range(len(result)):
					if steptype == "xpath":
						result[k] = result[k].xpath(instruction)[0]
					elif steptype == "prefix":
						result[k] = instruction + result[k]
					elif steptype == "suffix":
						result[k] = result[k] + instruction
					elif steptype == "rmprefix":
						if result[k].startswith(instruction): result[k] = result[k].replace(instruction,"",1)
					elif steptype == "regex":
						result[k] = re.sub(instruction,r"\1",result[k])
					elif steptype == "last":
						result[k] = result[k].split(instruction)[-1]

		i += 1

	return result

@defaultarguments(config,attempts="attempts")
def scrape(url,steps,requires_javascript=False,attempts=DEFAULT):
	"""Scrapes the given URL with the supplied steps and returns the result.

	:param string url: URL to scrape
	:param list steps: Steps to apply to the top node of the document
	:param bool requires_javascript: Whether the page as it displays in a js-enabled browser should be scraped or merely the source
	:return: Resulting node or string
	"""

	#print("Scraping",url)

	if requires_javascript:
		tree = _scrape_selenium(url)
	else:
		tree = _scrape_soup(url,attempts=attempts)


	return parse(tree,steps)


class _GoodException(Exception):
	pass

@defaultarguments(config,attempts="attempts")
def scrape_all(base_url,steps_elements,steps_content,start_page=0,page_multiplier=1,stop=math.inf,stopif={},attempts=DEFAULT,requires_javascript=False):
	"""Function to scrape a library, gallery or feed that consists of well-patterned elements and return all elements
	represented by the specified attributes and contents.

	:param string base_url: The URL part all pages have in common, with {page} as a placeholder for the page
	:param list steps_elements: Steps from the document node to get a list of all elements to be scraped
	:param dict steps_content: For each desired information about each element, a list of steps to reach this information from the element node
	:param string/int start_page: The part that must be supplied to the URL to get the first page
	:param int/func page_multiplier: Either an integer that the page needs to be multiplied with to get the correct URL, or a function that takes the page number as argument
	:param int stop: Limit of elements to scrape
	:param dict stopif: For any element attribute, a function that evaluates to True if scraping should be stopped
	:param int attempts: How many times a page visit should be attempted
	:return: A list of all elements as dictionaries of their attributes

	"""

	# set the page function
	if isinstance(page_multiplier,int):
		getpage = lambda x: x*page_multiplier
	else:
		getpage = page_multiplier


	pagenum = 0
	returned = 0

	try:
		while returned < stop:
			url = base_url.format(page=getpage(pagenum))

			#print("Page",pagenum,"URL",url)
			elements = scrape(url,steps_elements,attempts=attempts,requires_javascript=requires_javascript)
			#print(len(elements),"on this page")

			for e in elements:
				if returned >= stop: raise _GoodException("Number of elements reached")
				resultelement = {}

				try:
					for attribute in steps_content:
						#print("Attribute",attribute)
						res = parse(e,steps_content[attribute])
						#print("Setting attribute",attribute,"to",res)
						if attribute in stopif and stopif[attribute](res):
							raise _GoodException("Break condition reached")
						resultelement[attribute] = res

					#result.append(resultelement)
					yield resultelement
					returned += 1
				except _GoodException:
					raise
				except:
					#print("Failed to get element, skipping...")
					pass

			pagenum += 1
			if url == base_url.format(page=getpage(pagenum)): break

	except _GoodException:
		pass
	except:
		raise

	return






def _scrape_soup(url,attempts):
	br = mechanicalsoup.StatefulBrowser()
	for attempt in range(attempts):
		try:
			page = br.open(url)
			tree = html.fromstring(page.content)
			return tree
		except:
			print("Problem while scraping",url)
			time.sleep(1 + attempt)



def _scrape_selenium(url):
	try:
		from selenium import webdriver
		import time
		dr = webdriver.Firefox()
	except:
		print("Selenium and the Firefox web driver are needed to scrape sites with javascript")
		return None

	dr.get(url)
	oldtree = None
	for x in range(10):
		dr.execute_script("window.scrollTo(0, 10000);")
		time.sleep(1)
		raw = dr.execute_script("return document.documentElement.outerHTML")
		tree = html.fromstring(raw)
		if dr.execute_script("(window.innerHeight + window.scrollY) >= document.body.offsetHeight"):
			break
	dr.close()
	return tree



import setuptools
import importlib

module = importlib.import_module(setuptools.find_packages()[0])

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=module.name,
    version=".".join(str(n) for n in module.version),
    author=module.author["name"],
    author_email=module.author["email"],
    description="Small toolkit of utilities for python projects",
	license="GPLv3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/" + module.author["github"] + "/" + module.name,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
	python_requires=">=3.5",
	install_requires=module.requires,
	package_data={'': ['data/login.pyhp']},
	include_package_data=True,
)
